/*******************************************************************************

NAME            networkFilter.cc
VERSION         %version:  1 %
UPDATE DATE     %date_modified: Tue Oct 20 12:28:31 2015 %
PROGRAMMER      %created_by:    bradfort %

        Copyright 2015 NetApp, Inc. All Rights Reserved.

DESCRIPTION: Daemon to check to intercept IPv6 router advertisements and check
             for reserved address conflicts

*******************************************************************************/

/**
 * \file networkFilter.cc
 *
 * Daemon process to intercept IPv6 router advertisements and check for
 * reserved address conflicts.  Drop conflicting packets.
 */

#include <stdlib.h>
#include <stddef.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <time.h>
#include <errno.h>
#include <limits.h>
#include <netinet/in.h>
#include <netdb.h>
#include <ctype.h>
#include <systemd/sd-journal.h>
#include <sys/socket.h>
#include <string>

#include "networkFilter.h"

/** Globals **/
/**
 * Display usage
 *
 * \param process  Name of the process
 */
void
usage
    (
    const char* process
    )
{
    fprintf(stderr, "%s [options]\n"
                    "  -d    Run as Daemon\n"
                    "  -h    Help\n"
                    "  -v    Verbose\n",
                    process
          );
}

/**
 * Compare bits for two character arrays
 *
 * \param arr1  First array
 * \param arr2  Second array
 * \param nBits Number of bits to compare
 *
 * \return true of bits compare
 *
 * Assumes bits start with MS bit of each array
 */
bool
NetworkFilter::compareBits
    (
    uint8_t* arr1,
    uint8_t* arr2,
    uint32_t nBits
    )
{
    uint32_t nBytes = nBits / CHAR_BIT;
    uint32_t rem    = nBits % CHAR_BIT;

    bool result = (memcmp(arr1, arr2, nBytes) == 0);

    // compare any remaining bits
    if (result && rem)
    {
        //  Bits Remaining:         0     1     2     3     4     5     6     7
        static uint8_t mask[] = {0x00, 0x80, 0xC0, 0xE0, 0xF0, 0xF8, 0xFC, 0xFE};

        result = (((arr1[nBytes] ^ arr2[nBytes] ) & mask[rem]) == 0);
    }
    return result;
}

/**
 * Walk through the Router Advertisement records
 *
 * \param pktData  Pointer to the Data packet
 * \param len      The length of the data packet
 *
 * \return true if there is an IP match
 */
bool
NetworkFilter::findPrefixIP
    (
    uint8_t* pktData,
    int      len
    )
{
    uint32_t nbits;
    IpPacket* pPkt = reinterpret_cast<struct IpPacket*>(pktData);
    uint8_t*  pOpt = pPkt->options;

    // Validate the RA header
    if (pPkt->ra.type != RA)
        return false;

    len -= offsetof(IpPacket, options);
    while (len > 0)
    {
        std::string    ip;
        uint8_t*       puc;
        struct Prefix* pPrefix = reinterpret_cast<struct Prefix*>(pOpt);
        char           str[4];

        // If the option is Prefix
        if (pPrefix->type == PREFIX)
        {
            if (m_Verbose)
                printf("Prefix\n");
            nbits = pPrefix->prefixLength;
            if (nbits > 48)
                nbits = 48;
            if (nbits < 16)
               return false;

            puc = pPrefix->prefixIP;

            for (size_t i = 0; i < sizeof(pPrefix->prefixIP); i++)
            {
                snprintf(str, sizeof(str),"%02x", puc[i]);
                ip += str;
                if ((i & 1) && (i + 1 < sizeof(pPrefix->prefixIP)))
                    ip += ":";
            }
            if (m_Verbose)
                printf("PrefixIP=%s\n", ip.c_str());

            if (compareBits(m_ThisPrefix, puc, nbits))
            {
                sd_journal_print(LOG_ERR, "Warning: dropping Router Advertisement for %s",
                        ip.c_str());
                if (m_Verbose)
                    printf("Warning: dropping Router Advertisement for %s\n", ip.c_str());
                return true;
            }
            return false;
        }

        uint32_t optionLength = pPrefix->length * 8;

        // If not Prefix, then just move along
        if (m_Verbose)
            printf("NetworkFilter: Router Advertisement option type=%d length=%d\n",
                   pPrefix->type, optionLength);
        sd_journal_print(LOG_DEBUG, "Router Advertisement option type=%d length=%d",
                pPrefix->type, optionLength);
        pOpt += optionLength;
        len  -= optionLength;
    }

    return false;
}

/**
 * Network Filter callback function C interface
 *
 * \param myQueue  Pointer to network filter queue handle
 * \param msg      Pointer to generated message
 * \param pkt      Pointer to the network packet
 * \param cbData   Pointer to Call back data .. in this case
 *                 needs to be the pointer to the NetworkFilter class
 *
 * \return status from setting the verdict
 */
int
NetworkFilter::callBack
    (
    nfq_q_handle*    myQueue,
    struct nfgenmsg* /* msg */,
    nfq_data*        pkt,
    void*            cbData
    )
{
    NetworkFilter* pnf = reinterpret_cast<NetworkFilter*>(cbData);

    if (pnf)
        return pnf->callBack(myQueue, pkt);
    return -1;
}

/**
 * Network Filter callback function to determine the verdict
 *
 * \param myQueue  Pointer to network filter queue handle
 * \param msg      Pointer to generated message
 * \param pkt      Pointer to the network packet
 * \param cbData   Pointer to Call back data .. in this case
 *                 needs to be the pointer to the NetworkFilter class
 *
 * \return status from setting the verdict
 */
int
NetworkFilter::callBack
    (
    nfq_q_handle*    myQueue,
    nfq_data*        pkt
    )
{
    uint32_t id = 0;
    nfqnl_msg_packet_hdr* header;
    char                  buffer[128];
    std::string           report("Packet received: ");
    uint32_t verdict = NF_ACCEPT;

    header = nfq_get_msg_packet_hdr(pkt);
    if (header)
    {
        id = ntohl(header->packet_id);

        snprintf(buffer, sizeof(buffer), "id %u; hw_protocol 0x%04x; hook %c ; ",
                 id, ntohs(header->hw_protocol),
                 ('0' + header->hook));
        report += buffer;
    }

    nfqnl_msg_packet_hw* macAddr = nfq_get_packet_hw(pkt);
    if (macAddr)
    {
        snprintf(buffer, sizeof(buffer), "mac len %d addr 0x", ntohl(macAddr->hw_addrlen));
        report += buffer;
        for (int i = 0; i < 8; i++)
        {
            snprintf(buffer, sizeof(buffer),"%02x", macAddr->hw_addr[i]);
            report += buffer;
        }
    }
    else
    {
        report += "no MAC addr";
    }

    timeval tv;

    if (!nfq_get_timestamp(pkt, &tv))
    {
        snprintf(buffer, sizeof(buffer), "; tstamp %ld.%ld", tv.tv_sec, tv.tv_usec);
        report += buffer;
    }
    else
    {
        report += "; no tstamp";
    }

    snprintf(buffer, sizeof(buffer), "; mark %u", nfq_get_nfmark(pkt));
    report += buffer;

    snprintf(buffer, sizeof(buffer), "; indev %u", nfq_get_indev(pkt));
    report += buffer;
    snprintf(buffer, sizeof(buffer), "; outdef %u", nfq_get_outdev(pkt));
    report += buffer;

    if (m_Verbose)
        printf("%s\n", report.c_str());
    sd_journal_print(LOG_DEBUG, "%s", report.c_str());

    // payload -- this  pk
    uint8_t* pktData;
    int      len = nfq_get_payload(pkt, &pktData);

    if (len > 0)
    {
        if (m_Verbose)
            printf("data[%d]:\n", len);
        sd_journal_print(LOG_DEBUG, "data[%d]:", len);
        report = "";
        // print in rows of 16
        for (int i = 0; i < len; i++)
        {
            snprintf(buffer, sizeof(buffer), "%02x ", pktData[i]);
            report += buffer;
            if (((i + 1) & 15) == 0)
            {
                if (m_Verbose)
                    printf(" %s\n", report.c_str());
                sd_journal_print(LOG_DEBUG, "%s", report.c_str());
                report = "";
            }
        }

        if (report.length())
        {
            if (m_Verbose)
                printf(" %s\n", report.c_str());
            sd_journal_print(LOG_DEBUG, "%s", report.c_str());
        }

        // Walk through packet to find the Prefix IP
        if (findPrefixIP(pktData, len))
            verdict = NF_DROP;
    }

    return nfq_set_verdict(myQueue, id, verdict, 0, 0);
}

/**
 * \param saddr points to a sockaddr struct to copy
 *        the address to.
 * \param name is a domain name to obtain an address for.
 * \returns zero on success or an error value consistent with
 *        the function getaddrinfo() or -1.
 *
 * This method attempts to resolve a node \e name into an IP
 * address.  If \e name is a dotted numerical address, then
 * the binary representation of it is copied into the address
 * field of saddr.
 */
int
NetworkFilter::getAddress
    (
    sockaddr_in6* saddr,
    const char* name
    )
{
    struct addrinfo *resolved = 0;
    struct addrinfo hints = {
        0,        // ai_flags
        AF_INET6, // ai_family (Only get IPv6 addresses)
        0,        // ai_socktype
        0,        // ai_protocol
        0,        // ai_addrlen
        0,        // ai_addr
        0,        // ai_canonname
        0         // ai_next
    };

    int err = getaddrinfo( name, 0, &hints, &resolved );

    if ( err )
    {
        fprintf(stderr, "Obtaining address for '%s'\n", name );
        return err;
    }

    // resolved must not be null
    if ( !resolved )
    {
        fprintf(stderr, "getAddress: resolved address points to zero.\n");
        return -1;
    }

    sockaddr_in6* resolvedAddr =
        reinterpret_cast<sockaddr_in6*>(resolved->ai_addr);
    memcpy(&saddr->sin6_addr, &resolvedAddr->sin6_addr,
           sizeof(saddr->sin6_addr) );

    // Release the address information memory.
    freeaddrinfo( resolved );

    return 0;
}

/**
 * Constructor for Network filter class
 *
 * \param verbose  Flag to set the verbosity
 *
 * \throw  Network Filter setup errors
 */
NetworkFilter::NetworkFilter
    (
    bool verbose
    ) : m_Verbose(verbose),
        m_NfqHandle(0),
        m_Queue(0)
{
    // Get the host information for this controller
    sockaddr_in6 saddr;

    memset(m_ThisPrefix, 0, sizeof(m_ThisPrefix));
    if (getAddress(&saddr, "dom0-this"))
    {
        throw NetworkFilterGetAddressException();
    }
    memcpy(m_ThisPrefix, saddr.sin6_addr.s6_addr, sizeof(m_ThisPrefix));

    if (m_Verbose)
    {
        printf("m_ThisPrefix = ");
        for (uint32_t i = 0; i < sizeof(m_ThisPrefix); i++)
            printf("%02x", m_ThisPrefix[i]);
        printf("\n");
    }

    // Get a queue connection handle
    m_NfqHandle = nfq_open();
    if (!m_NfqHandle)
    {
        throw NetworkFilterHandleException();
    }

    // unbind the handler from processing any IP packets
    if (nfq_unbind_pf(m_NfqHandle, AF_INET) < 0)
    {
        throw NetworkFilterUnbindException();
    }

    // Bind this handler to process IP packets
    if (nfq_bind_pf(m_NfqHandle, AF_INET) < 0)
    {
        throw NetworkFilterBindException();
    }

    // Install a callback on queue 0
    m_Queue = nfq_create_queue(m_NfqHandle, 0, callBack, this);
    if (!m_Queue)
    {
        throw NetworkFilterQueueCreateException();
    }

    // Turn on packet copy mode
    if (nfq_set_mode(m_Queue, NFQNL_COPY_PACKET, 0xffff) < 0)
    {
        throw NetworkFilterQueueCopyException();
    }
}

/**
 * Read messages
 *
 * \return  this function is not expected to return.
 */
void
NetworkFilter::readMessages
    (
    )
{
    int    res;
    char   buf[256];

    // derive a file descriptor so that we can read packets
    struct nfnl_handle* netlinkHandle = nfq_nfnlh(m_NfqHandle);
    int                 fd            = nfnl_fd(netlinkHandle);

    // Read packets until error
    while ((res = recv(fd, buf, sizeof(buf), 0)) && (res >= 0))
    {
        // If a packet was received, trigger callback
        nfq_handle_packet(m_NfqHandle, buf, res);
    }

    sd_journal_print(LOG_INFO, "NetworkFilter recv() returned %d errno=%d(%s)",
            res, errno, strerror(errno));
    fprintf(stderr, "NetworkFilter recv() returned %d errno=%d(%s)",
            res, errno, strerror(errno));
}

/**
 * Destructor for Network Filter class
 *
 */
NetworkFilter::~NetworkFilter
    (
    )
{
    // tear down
    nfq_destroy_queue(m_Queue);
    nfq_close(m_NfqHandle);
}

/**
 * Filter the Queued messages
 *
 * \param verbose   -- The verbose option
 */
void
filterMessages
    (
    bool verbose
    )
{
    NetworkFilter nF(verbose);

    // Read messages forever or until error
    nF.readMessages();
}

/**
 * Process main entry
 *
 * \param argc  The number of command line arguments
 * \param arvgv Pointer to arguments strings
 *
 * \return 0 on succes or 1 on error
 */
int
main
    (
    int argc,
    char** argv
    )
{
    bool           verbose    = false;
    bool           background = false;
    int            ret        = 0;
    int            c;

    // Parse the command line for options
    while ((c = getopt (argc, argv, "dhv?")) != -1)
    {
        switch (c)
        {
        case 'd':
            background = true;
            break;

        case '?':  /* Fall thru */
        case 'h':
            usage(argv[0]);
            exit(0);

        case 'v':
            verbose = true;
            break;

        default:
            usage(argv[0]);
            exit(-1);
        }
    }

    if (verbose)
        printf("%s Entry....\n", argv[0]);

    sd_journal_print(LOG_DEBUG, "%s Entry....", argv[0]);

    if (background)
    {
        // Fork a new process and go background.
        pid_t pid = fork();
        if (pid < 0)
        {
            sd_journal_print(LOG_ERR, "Fork failed. %s", strerror(errno));
            return -1;
        }

        // Exit the parent process.
        if (pid > 0)
            return 0;
    }

    // Create Filter class and filter Queued messages
    try
    {
        filterMessages(verbose);
    }
    catch(std::exception& ex)
    {
        fprintf(stderr, "Error (%s) with NetworkFilter class \n", ex.what());
        sd_journal_print(LOG_ERR, "Error (%s) with NetworkFilter class", ex.what());
        ret = -1;
    }


    sd_journal_print(LOG_DEBUG, "%s Exiting....", argv[0]);
    if (verbose)
        printf("%s Exiting....", argv[0]);

    return ret;
}
