/*******************************************************************************

NAME            networkFilter.h
VERSION         %version:  2 %
UPDATE DATE     %date_modified: Thu Mar 30 10:06:47 2017 %
PROGRAMMER      %created_by:    dhoyer %

        Copyright 2015-2017 NetApp, Inc. All Rights Reserved.

*******************************************************************************/

/**
 * \file networkFilter.h
 *
 * Define classes and structures needed for filtering
 * Advertisement messages
 */

#ifndef __INCnetworkFilter
#define __INCnetworkFilter
#include <stdint.h>
#include <exception>
#include <eseries/cmnTypes.h>

extern "C"
{
#include <linux/netfilter.h>
#include <libnetfilter_queue/libnetfilter_queue.h>
}

/**
 * Structure for Router Advertisement Header
 */
struct RouterAdv
{
    UINT8    type;
    UINT8    code;
    UINT8    checksum[2];
    UINT8    hopLimit;
    UINT8    flags;
    UINT8    routerLifetime[2];
    UINT8    reachableTime[4];
    UINT8    retransTimer[4];
} ;


/**
 * IP packet containting a Router Advertisement Record
 */

struct IpPacket
{
    UINT8     versionTrafficFlowLabel[4];
    UINT8     payloadLength[2];
    UINT8     nextHeader;
    UINT8     hopLimit;
    UINT8     sourceIP[16];
    UINT8     destIP[16];
    struct RouterAdv ra;
    UINT8     options[];   // RA options start here
} ;

/**
 * Structure for Prefix option
 */
struct Prefix
{
    UINT8    type;
    UINT8    length;
    UINT8    prefixLength;
    UINT8    lar;
    UINT8    validLifeTime[4];
    UINT8    preferredLifeTime[4];
    UINT8    reserved[4];
    UINT8    prefixIP[16];
} ;


/**
 * This class manages the Accepting and dropping of IP Advertisement messages
 */
class NetworkFilter
{
public:

    enum ADV_TYPES {LINK_LAYER = 1, PREFIX = 3, RA = 134};

    // Constructor
    NetworkFilter(bool verbose);

    // Destructor
    ~NetworkFilter();

    // Read message packets
    void readMessages();

private:

    // get address for a given host name
    int getAddress(sockaddr_in6* saddr, const char* name);

    // Compare bits for two arrays
    bool compareBits (UINT8* arr1, UINT8* arr2, uint32_t nBits);

    // Find the PrefixIP out of the provided Advertisement messages
    bool findPrefixIP(UINT8* pktData, int len);

    // C-Linkange to the Network Filter Callback funtion
    static int callBack(nfq_q_handle* myQueue, struct nfgenmsg*  msg,
                        nfq_data* pkt, void* cbData);

    // Network Filter Callback function
    int callBack(nfq_q_handle* myQueue, nfq_data* pkt);

    bool     m_Verbose;                //< Verbosity flag
    struct   nfq_handle* m_NfqHandle;  //< Handle for nfq functions
    struct   nfq_q_handle* m_Queue;    //< Pointer to network filter queue
    UINT8    m_ThisPrefix[16];         //< IP for this Controller
};

/**
 * NetworkFilter Exceptions
 *
 * \{
 */
class NetworkFilterGetAddressException: public std::exception
{
    virtual const char* what() const throw()
    {
        return "Unable to resolve address for dom0-this";
    }
};

class NetworkFilterHandleException: public std::exception
{
    virtual const char* what() const throw()
    {
        return "Error opening nfq handle";
    }
};

class NetworkFilterUnbindException: public std::exception
{
    virtual const char* what() const throw()
    {
        return "Error attempting to unbind NFQ handler";
    }
};

class NetworkFilterBindException: public std::exception
{
    virtual const char* what() const throw()
    {
        return "Error attempting to bind NFQ handler";
    }
};

class NetworkFilterQueueCreateException: public std::exception
{
    virtual const char* what() const throw()
    {
        return "Error attempting to create NFQ Queue";
    }
};

class NetworkFilterQueueCopyException: public std::exception
{
    virtual const char* what() const throw()
    {
        return "Error attempting to set NFQ Queue mode";
    }
};

/**
 * \}
 */


#endif        /* End of __INCnetwork-filter */
